@@ -63,6 +63,17 @@ static struct gateway_data *find_gateway(int index, const char *gateway)
 	return NULL;
 }
 
+static void remove_gateway(int index, const char *gateway)
+{
+	struct gateway_data *data;
+
+	data = find_gateway(index, gateway);
+	if (data == NULL)
+		return;
+
+	gateway_list = g_slist_remove(gateway_list, data);
+}
+
 static int set_route(struct connman_element *element, const char *gateway)
 {
 	struct ifreq ifr;
@@ -338,12 +349,12 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_variant(&dict, "Default",
 					DBUS_TYPE_BOOLEAN, &element->enabled);
 
-	device = __connman_element_get_device(element);
+	device = __connman_element_get_device_path(element);
 	if (device != NULL)
 		connman_dbus_dict_append_variant(&dict, "Device",
 					DBUS_TYPE_OBJECT_PATH, &device);
 
-	network = __connman_element_get_network(element);
+	network = __connman_element_get_network_path(element);
 	if (network != NULL)
 		connman_dbus_dict_append_variant(&dict, "Network",
 					DBUS_TYPE_OBJECT_PATH, &network);
@@ -456,6 +467,7 @@ static void unregister_interface(struct connman_element *element)
 
 static int connection_probe(struct connman_element *element)
 {
+	struct connman_service *service;
 	const char *gateway = NULL;
 
 	DBG("element %p name %s", element, element->name);
@@ -474,6 +486,9 @@ static int connection_probe(struct connman_element *element)
 	if (register_interface(element) < 0)
 		return -ENODEV;
 
+	service = __connman_element_get_service(element);
+	__connman_service_ready(service);
+
 	if (gateway == NULL)
 		return 0;
 
@@ -498,10 +513,14 @@ done:
 
 static void connection_remove(struct connman_element *element)
 {
+	struct connman_service *service;
 	const char *gateway = NULL;
 
 	DBG("element %p name %s", element, element->name);
 
+	service = __connman_element_get_service(element);
+	__connman_service_disconnect(service);
+
 	unregister_interface(element);
 
 	connman_element_get_value(element,
@@ -512,6 +531,8 @@ static void connection_remove(struct connman_element *element)
 	if (gateway == NULL)
 		return;
 
+	remove_gateway(element->index, gateway);
+
 	connman_element_set_enabled(element, FALSE);
 	emit_default_signal(element);
 
