@@ -24,6 +24,8 @@
 #endif
 
 #include <errno.h>
+#include <string.h>
+
 #include <gdbus.h>
 
 #include "connman.h"
@@ -33,23 +35,48 @@ struct connman_device {
 	enum connman_device_type type;
 	enum connman_device_mode mode;
 	enum connman_device_policy policy;
+	connman_bool_t secondary;
 	connman_bool_t powered;
 	connman_bool_t carrier;
 	connman_bool_t scanning;
+	connman_bool_t disconnected;
 	connman_uint8_t priority;
+	connman_uint16_t scan_interval;
 	char *name;
 	char *node;
+	char *address;
 	char *interface;
+	char *ident;
 	unsigned int connections;
+	guint scan_timeout;
 
 	struct connman_device_driver *driver;
 	void *driver_data;
 
 	connman_bool_t registered;
 
+	char *last_network;
+	struct connman_network *network;
 	GHashTable *networks;
 };
 
+static gboolean device_scan_trigger(gpointer user_data)
+{
+	struct connman_device *device = user_data;
+
+	DBG("device %p", device);
+
+	if (device->driver == NULL) {
+		device->scan_timeout = 0;
+		return FALSE;
+	}
+
+	if (device->driver->scan)
+		device->driver->scan(device);
+
+	return TRUE;
+}
+
 static const char *type2description(enum connman_device_type type)
 {
 	switch (type) {
@@ -64,6 +91,8 @@ static const char *type2description(enum connman_device_type type)
 		return "WiMAX";
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		return "Bluetooth";
+	case CONNMAN_DEVICE_TYPE_GPS:
+		return "GPS";
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
@@ -88,6 +117,8 @@ static const char *type2string(enum connman_device_type type)
 		return "wimax";
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		return "bluetooth";
+	case CONNMAN_DEVICE_TYPE_GPS:
+		return "gps";
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
@@ -130,6 +161,60 @@ static enum connman_device_policy string2policy(const char *policy)
 		return CONNMAN_DEVICE_POLICY_UNKNOWN;
 }
 
+static int set_carrier(struct connman_device *device, connman_bool_t carrier)
+{
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_device(device);
+	__connman_service_set_carrier(service, carrier);
+
+	if (carrier == TRUE) {
+		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
+		struct connman_element *element;
+
+		device->disconnected = TRUE;
+
+		switch (device->policy) {
+		case CONNMAN_DEVICE_POLICY_UNKNOWN:
+		case CONNMAN_DEVICE_POLICY_IGNORE:
+		case CONNMAN_DEVICE_POLICY_OFF:
+		case CONNMAN_DEVICE_POLICY_MANUAL:
+			return 0;
+		case CONNMAN_DEVICE_POLICY_AUTO:
+			break;
+		}
+
+		switch (device->element.ipv4.method) {
+		case CONNMAN_IPV4_METHOD_UNKNOWN:
+		case CONNMAN_IPV4_METHOD_OFF:
+			return 0;
+		case CONNMAN_IPV4_METHOD_STATIC:
+			type = CONNMAN_ELEMENT_TYPE_IPV4;
+			break;
+		case CONNMAN_IPV4_METHOD_DHCP:
+			type = CONNMAN_ELEMENT_TYPE_DHCP;
+			break;
+		}
+
+		element = connman_element_create(NULL);
+		if (element != NULL) {
+			element->type  = type;
+			element->index = device->element.index;
+
+			if (connman_element_register(element,
+							&device->element) < 0)
+				connman_element_unref(element);
+
+			device->disconnected = FALSE;
+
+			__connman_service_indicate_configuration(service);
+		}
+	} else
+		connman_element_unregister_children(&device->element);
+
+	return 0;
+}
+
 static int set_powered(struct connman_device *device, connman_bool_t powered)
 {
 	struct connman_device_driver *driver = device->driver;
@@ -141,16 +226,20 @@ static int set_powered(struct connman_device *device, connman_bool_t powered)
 		return -EINVAL;
 
 	if (powered == TRUE) {
-		if (driver->enable)
+		if (driver->enable) {
 			err = driver->enable(device);
-		else
+			__connman_notifier_device_type_increase(device->type);
+		} else
 			err = -EINVAL;
 	} else {
 		g_hash_table_remove_all(device->networks);
 
-		if (driver->disable)
+		set_carrier(device, FALSE);
+
+		if (driver->disable) {
 			err = driver->disable(device);
-		else
+			__connman_notifier_device_type_decrease(device->type);
+		} else
 			err = -EINVAL;
 	}
 
@@ -184,6 +273,8 @@ static int set_policy(DBusConnection *connection,
 	case CONNMAN_DEVICE_POLICY_MANUAL:
 		if (device->powered == FALSE)
 			err = set_powered(device, TRUE);
+		else
+			err = set_carrier(device, device->carrier);
 		break;
 	}
 
@@ -213,6 +304,15 @@ static int set_policy(DBusConnection *connection,
 	return 0;
 }
 
+static void append_path(gpointer key, gpointer value, gpointer user_data)
+{
+	struct connman_element *element = value;
+	DBusMessageIter *iter = user_data;
+
+	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
+							&element->path);
+}
+
 static void append_networks(struct connman_device *device,
 						DBusMessageIter *entry)
 {
@@ -227,8 +327,7 @@ static void append_networks(struct connman_device *device,
 
 	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
 				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_element_list((struct connman_element *) device,
-					CONNMAN_ELEMENT_TYPE_NETWORK, &iter);
+	g_hash_table_foreach(device->networks, append_path, &iter);
 	dbus_message_iter_close_container(&value, &iter);
 
 	dbus_message_iter_close_container(entry, &value);
@@ -259,20 +358,19 @@ static DBusMessage *get_properties(DBusConnection *conn,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	str = type2description(device->type);
-	if (str != NULL && device->interface != NULL) {
-		char *name = g_strdup_printf("%s (%s)", str, device->interface);
-		if (name != NULL)
-			connman_dbus_dict_append_variant(&dict, "Name",
-						DBUS_TYPE_STRING, &name);
-		g_free(name);
-	}
+	if (device->name != NULL)
+		connman_dbus_dict_append_variant(&dict, "Name",
+					DBUS_TYPE_STRING, &device->name);
 
 	str = type2string(device->type);
 	if (str != NULL)
 		connman_dbus_dict_append_variant(&dict, "Type",
 						DBUS_TYPE_STRING, &str);
 
+	if (device->address != NULL)
+		connman_dbus_dict_append_variant(&dict, "Address",
+					DBUS_TYPE_STRING, &device->address);
+
 	if (device->interface != NULL)
 		connman_dbus_dict_append_variant(&dict, "Interface",
 					DBUS_TYPE_STRING, &device->interface);
@@ -295,10 +393,16 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
+		break;
 	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		__connman_element_append_ipv4(&device->element, &dict);
 		break;
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		if (device->scan_interval > 0)
+			connman_dbus_dict_append_variant(&dict, "ScanInterval",
+				DBUS_TYPE_UINT16, &device->scan_interval);
+
 		dbus_message_iter_open_container(&dict, DBUS_TYPE_DICT_ENTRY,
 								NULL, &entry);
 		append_networks(device, &entry);
@@ -317,6 +421,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 	struct connman_device *device = data;
 	DBusMessageIter iter, value;
 	const char *name;
+	int type;
 
 	DBG("conn %p", conn);
 
@@ -331,10 +436,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
+	type = dbus_message_iter_get_arg_type(&value);
+
 	if (g_str_equal(name, "Powered") == TRUE) {
 		connman_bool_t powered;
 		int err;
 
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &powered);
 
 		if (device->powered == powered)
@@ -348,6 +458,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		const char *str;
 		int err;
 
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &str);
 		policy = string2policy(str);
 		if (policy == CONNMAN_DEVICE_POLICY_UNKNOWN)
@@ -359,9 +472,52 @@ static DBusMessage *set_property(DBusConnection *conn,
 	} else if (g_str_equal(name, "Priority") == TRUE) {
 		connman_uint8_t priority;
 
+		if (type != DBUS_TYPE_BYTE)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &priority);
 
 		device->priority = priority;
+	} else if (g_str_equal(name, "ScanInterval") == TRUE) {
+		connman_uint16_t interval;
+
+		switch (device->mode) {
+		case CONNMAN_DEVICE_MODE_UNKNOWN:
+		case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+			return __connman_error_invalid_arguments(msg);
+		case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+		case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+			break;
+		}
+
+		if (type != DBUS_TYPE_UINT16)
+			return __connman_error_invalid_arguments(msg);
+
+		dbus_message_iter_get_basic(&value, &interval);
+
+		device->scan_interval = interval;
+
+		if (device->scan_timeout > 0) {
+			g_source_remove(device->scan_timeout);
+			device->scan_timeout = 0;
+		}
+
+		if (device->scan_interval > 0) {
+			guint interval = device->scan_interval;
+			device->scan_timeout = g_timeout_add_seconds(interval,
+						device_scan_trigger, device);
+		}
+	} else if (g_str_has_prefix(name, "IPv4") == TRUE) {
+		switch (device->mode) {
+		case CONNMAN_DEVICE_MODE_UNKNOWN:
+		case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+		case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+			return __connman_error_invalid_arguments(msg);
+		case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+			__connman_element_set_ipv4(&device->element,
+								name, &value);
+			break;
+		}
 	}
 
 	__connman_storage_save_device(device);
@@ -369,6 +525,78 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *join_network(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct connman_device *device = data;
+	struct connman_network *network;
+	enum connman_network_type type;
+	DBusMessageIter iter, array;
+	int err, index;
+
+	DBG("conn %p", conn);
+
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	if (!device->driver || !device->driver->join)
+		return __connman_error_not_supported(msg);
+
+	dbus_message_iter_init(msg, &iter);
+	dbus_message_iter_recurse(&iter, &array);
+
+	switch (device->type) {
+	case CONNMAN_DEVICE_TYPE_WIFI:
+		type = CONNMAN_NETWORK_TYPE_WIFI;
+		break;
+	default:
+		return __connman_error_not_supported(msg);
+	}
+
+	network = connman_network_create("00_00_00_00_00_00", type);
+	if (network == NULL)
+		return __connman_error_failed(msg);
+
+	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry, value;
+		const char *key, *str;
+
+		dbus_message_iter_recurse(&array, &entry);
+		dbus_message_iter_get_basic(&entry, &key);
+
+		dbus_message_iter_next(&entry);
+		dbus_message_iter_recurse(&entry, &value);
+
+		switch (dbus_message_iter_get_arg_type(&value)) {
+		case DBUS_TYPE_STRING:
+			dbus_message_iter_get_basic(&value, &str);
+			if (g_str_equal(key, "WiFi.SSID") == TRUE)
+				connman_network_set_blob(network, key,
+							str, strlen(str));
+			else
+				connman_network_set_string(network, key, str);
+			break;
+		}
+
+		dbus_message_iter_next(&array);
+	}
+
+	index = connman_device_get_index(device);
+	connman_network_set_index(network, index);
+
+	connman_network_set_protocol(network, CONNMAN_NETWORK_PROTOCOL_IP);
+
+	err = device->driver->join(device, network);
+
+	connman_network_unref(network);
+
+	if (err < 0)
+		return __connman_error_failed(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *create_network(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -426,6 +654,7 @@ static DBusMessage *propose_scan(DBusConnection *conn,
 static GDBusMethodTable device_methods[] = {
 	{ "GetProperties", "",      "a{sv}", get_properties },
 	{ "SetProperty",   "sv",    "",      set_property   },
+	{ "JoinNetwork",   "a{sv}", "",      join_network   },
 	{ "CreateNetwork", "a{sv}", "o",     create_network },
 	{ "RemoveNetwork", "o",     "",      remove_network },
 	{ "ProposeScan",   "",      "",      propose_scan   },
@@ -521,8 +750,10 @@ static void device_enable(struct connman_device *device)
 	if (device->powered == TRUE)
 		return;
 
-	if (device->driver->enable)
+	if (device->driver->enable) {
 		device->driver->enable(device);
+		__connman_notifier_device_type_increase(device->type);
+	}
 }
 
 static void device_disable(struct connman_device *device)
@@ -537,8 +768,10 @@ static void device_disable(struct connman_device *device)
 
 	g_hash_table_remove_all(device->networks);
 
-	if (device->driver->disable)
+	if (device->driver->disable) {
 		device->driver->disable(device);
+		__connman_notifier_device_type_decrease(device->type);
+	}
 }
 
 static int setup_device(struct connman_device *device)
@@ -555,6 +788,17 @@ static int setup_device(struct connman_device *device)
 		return err;
 	}
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		if (device->secondary == FALSE)
+			__connman_profile_add_device(device);
+		break;
+	}
+
 	device_enable(device);
 
 	return 0;
@@ -572,6 +816,9 @@ static void probe_driver(struct connman_element *element, gpointer user_data)
 	if (element->device->driver != NULL)
 		return;
 
+	if (driver->type != element->device->type)
+		return;
+
 	if (driver->probe(element->device) < 0)
 		return;
 
@@ -586,6 +833,17 @@ static void remove_device(struct connman_device *device)
 
 	device_disable(device);
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		if (device->secondary == FALSE)
+			__connman_profile_remove_device(device);
+		break;
+	}
+
 	unregister_interface(&device->element);
 
 	if (device->driver->remove)
@@ -679,10 +937,14 @@ static void device_destruct(struct connman_element *element)
 
 	DBG("element %p name %s", element, element->name);
 
+	g_free(device->ident);
 	g_free(device->node);
 	g_free(device->name);
+	g_free(device->address);
 	g_free(device->interface);
 
+	g_free(device->last_network);
+
 	g_hash_table_destroy(device->networks);
 	device->networks = NULL;
 }
@@ -710,44 +972,56 @@ struct connman_device *connman_device_create(const char *node,
 
 	DBG("device %p", device);
 
-	device->element.refcount = 1;
+	__connman_element_initialize(&device->element);
 
 	device->element.name = g_strdup(node);
 	device->element.type = CONNMAN_ELEMENT_TYPE_DEVICE;
-	device->element.index = -1;
 
 	device->element.device = device;
 	device->element.destruct = device_destruct;
 
 	str = type2string(type);
 	if (str != NULL)
-		connman_element_add_static_property(&device->element,
-					"Type", DBUS_TYPE_STRING, &str);
+		connman_element_set_string(&device->element,
+					CONNMAN_PROPERTY_ID_TYPE, str);
 
-	device->type   = type;
-	device->mode   = CONNMAN_DEVICE_MODE_UNKNOWN;
-	device->policy = CONNMAN_DEVICE_POLICY_AUTO;
+	device->element.ipv4.method = CONNMAN_IPV4_METHOD_DHCP;
+
+	device->type      = type;
+	device->name      = g_strdup(type2description(device->type));
+	device->mode      = CONNMAN_DEVICE_MODE_UNKNOWN;
+	device->policy    = CONNMAN_DEVICE_POLICY_AUTO;
+	device->secondary = FALSE;
 
 	switch (type) {
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
 		device->priority = 0;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_ETHERNET:
 	case CONNMAN_DEVICE_TYPE_WIFI:
 		device->priority = 100;
+		device->scan_interval = 300;
 		break;
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 		device->priority = 20;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		device->priority = 50;
+		device->scan_interval = 0;
+		break;
+	case CONNMAN_DEVICE_TYPE_GPS:
+		device->priority = 0;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
 	case CONNMAN_DEVICE_TYPE_NOVATEL:
 		device->priority = 60;
+		device->scan_interval = 0;
 		break;
 	}
 
@@ -782,6 +1056,22 @@ void connman_device_unref(struct connman_device *device)
 	connman_element_unref(&device->element);
 }
 
+const char *__connman_device_get_type(struct connman_device *device)
+{
+	return type2string(device->type);
+}
+
+/**
+ * connman_device_get_type:
+ * @device: device structure
+ *
+ * Get type of device
+ */
+enum connman_device_type connman_device_get_type(struct connman_device *device)
+{
+	return device->type;
+}
+
 /**
  * connman_device_get_name:
  * @device: device structure
@@ -842,6 +1132,13 @@ void connman_device_set_interface(struct connman_device *device,
 
 	g_free(device->interface);
 	device->interface = g_strdup(interface);
+
+	if (device->name == NULL) {
+		const char *str = type2description(device->type);
+		if (str != NULL && device->interface != NULL)
+			device->name = g_strdup_printf("%s (%s)", str,
+							device->interface);
+	}
 }
 
 /**
@@ -855,6 +1152,25 @@ const char *connman_device_get_interface(struct connman_device *device)
 	return device->interface;
 }
 
+/**
+ * connman_device_set_ident:
+ * @device: device structure
+ * @ident: unique identifier
+ *
+ * Set unique identifier of device
+ */
+void connman_device_set_ident(struct connman_device *device,
+							const char *ident)
+{
+	g_free(device->ident);
+	device->ident = g_strdup(ident);
+}
+
+const char *__connman_device_get_ident(struct connman_device *device)
+{
+	return device->ident;
+}
+
 /**
  * connman_device_set_policy:
  * @device: device structure
@@ -881,6 +1197,41 @@ void connman_device_set_mode(struct connman_device *device,
 	device->mode = mode;
 }
 
+/**
+ * connman_device_get_mode:
+ * @device: device structure
+ *
+ * Get network mode of device
+ */
+enum connman_device_mode connman_device_get_mode(struct connman_device *device)
+{
+	return device->mode;
+}
+
+/**
+ * connman_device_set_secondary:
+ * @device: device structure
+ * @secondary: secondary value
+ *
+ * Change secondary value of device
+ */
+void connman_device_set_secondary(struct connman_device *device,
+                                                connman_bool_t secondary)
+{
+	device->secondary = secondary;
+}
+
+/**
+ * connman_device_get_secondary:
+ * @device: device structure
+ *
+ * Get secondary value of device
+ */
+connman_bool_t connman_device_get_secondary(struct connman_device *device)
+{
+	return device->secondary;
+}
+
 /**
  * connman_device_set_powered:
  * @device: device structure
@@ -902,6 +1253,9 @@ int connman_device_set_powered(struct connman_device *device,
 
 	device->powered = powered;
 
+	if (device->registered == FALSE)
+		return 0;
+
 	signal = dbus_message_new_signal(device->element.path,
 				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
 	if (signal == NULL)
@@ -924,6 +1278,17 @@ int connman_device_set_powered(struct connman_device *device,
 	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
 		return 0;
 
+	if (device->scan_timeout > 0) {
+		g_source_remove(device->scan_timeout);
+		device->scan_timeout = 0;
+	}
+
+	if (device->scan_interval > 0) {
+		guint interval = device->scan_interval;
+		device->scan_timeout = g_timeout_add_seconds(interval,
+					device_scan_trigger, device);
+	}
+
 	if (device->driver->scan)
 		device->driver->scan(device);
 
@@ -956,20 +1321,36 @@ int connman_device_set_carrier(struct connman_device *device,
 
 	device->carrier = carrier;
 
-	if (carrier == TRUE) {
-		struct connman_element *element;
+	return set_carrier(device, device->carrier);
+}
 
-		element = connman_element_create(NULL);
-		if (element != NULL) {
-			element->type  = CONNMAN_ELEMENT_TYPE_DHCP;
-			element->index = device->element.index;
+int __connman_device_connect(struct connman_device *device)
+{
+	DBG("device %p", device);
 
-			if (connman_element_register(element,
-							&device->element) < 0)
-				connman_element_unref(element);
-		}
-	} else
-		connman_element_unregister_children(&device->element);
+	if (device->disconnected == FALSE)
+		return -EINVAL;
+
+	return 0;
+}
+
+int __connman_device_disconnect(struct connman_device *device)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+
+	DBG("device %p", device);
+
+	connman_device_set_disconnected(device, TRUE);
+
+	g_hash_table_iter_init(&iter, device->networks);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct connman_network *network = value;
+
+		if (connman_network_get_connected(network) == TRUE)
+			__connman_network_disconnect(network);
+	}
 
 	return 0;
 }
@@ -988,9 +1369,22 @@ static void connect_known_network(struct connman_device *device)
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		connman_uint8_t old_priority, new_priority;
 		connman_uint8_t old_strength, new_strength;
+		const char *name;
 
 		count++;
 
+		if (connman_network_get_available(value) == FALSE)
+			continue;
+
+		name = connman_network_get_string(value,
+						CONNMAN_PROPERTY_ID_NAME);
+		if (name != NULL && device->last_network != NULL) {
+			if (g_str_equal(name, device->last_network) == TRUE) {
+				network = value;
+				break;
+			}
+		}
+
 		if (connman_network_get_remember(value) == FALSE)
 			continue;
 
@@ -999,8 +1393,10 @@ static void connect_known_network(struct connman_device *device)
 			continue;
 		}
 
-		old_priority = connman_network_get_uint8(network, "Priority");
-		new_priority = connman_network_get_uint8(value, "Priority");
+		old_priority = connman_network_get_uint8(network,
+						CONNMAN_PROPERTY_ID_PRIORITY);
+		new_priority = connman_network_get_uint8(value,
+						CONNMAN_PROPERTY_ID_PRIORITY);
 
 		if (new_priority != old_priority) {
 			if (new_priority > old_priority)
@@ -1008,8 +1404,10 @@ static void connect_known_network(struct connman_device *device)
 			continue;
 		}
 
-		old_strength = connman_network_get_uint8(network, "Strength");
-		new_strength = connman_network_get_uint8(value, "Strength");
+		old_strength = connman_network_get_uint8(network,
+						CONNMAN_PROPERTY_ID_STRENGTH);
+		new_strength = connman_network_get_uint8(value,
+						CONNMAN_PROPERTY_ID_STRENGTH);
 
 		if (new_strength > old_strength)
 			network = value;
@@ -1030,6 +1428,34 @@ static void connect_known_network(struct connman_device *device)
 		device->driver->scan(device);
 }
 
+static void mark_network_unavailable(gpointer key, gpointer value,
+							gpointer user_data)
+{
+	struct connman_network *network = value;
+
+	if (connman_network_get_connected(network) == TRUE)
+		return;
+
+	connman_network_set_available(network, FALSE);
+}
+
+static gboolean remove_unavailable_network(gpointer key, gpointer value,
+							gpointer user_data)
+{
+	struct connman_network *network = value;
+
+	if (connman_network_get_connected(network) == TRUE)
+		return FALSE;
+
+	if (connman_network_get_remember(network) == TRUE)
+		return FALSE;
+
+	if (connman_network_get_available(network) == TRUE)
+		return FALSE;
+
+	return TRUE;
+}
+
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1070,12 +1496,32 @@ int connman_device_set_scanning(struct connman_device *device,
 
 	g_dbus_send_message(connection, signal);
 
-	if (scanning == TRUE)
+	if (scanning == TRUE) {
+		if (device->scan_timeout > 0) {
+			g_source_remove(device->scan_timeout);
+			device->scan_timeout = 0;
+		}
+
+		if (device->scan_interval > 0) {
+			guint interval = device->scan_interval;
+			device->scan_timeout = g_timeout_add_seconds(interval,
+						device_scan_trigger, device);
+		}
+
+		g_hash_table_foreach(device->networks,
+					mark_network_unavailable, NULL);
 		return 0;
+	}
+
+	g_hash_table_foreach_remove(device->networks,
+					remove_unavailable_network, NULL);
 
 	if (device->connections > 0)
 		return 0;
 
+	if (device->disconnected == TRUE)
+		return 0;
+
 	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
 		return 0;
 
@@ -1084,6 +1530,35 @@ int connman_device_set_scanning(struct connman_device *device,
 	return 0;
 }
 
+/**
+ * connman_device_set_disconnected:
+ * @device: device structure
+ * @disconnected: disconnected state
+ *
+ * Change disconnected state of device (only for device with networks)
+ */
+int connman_device_set_disconnected(struct connman_device *device,
+						connman_bool_t disconnected)
+{
+	DBG("driver %p disconnected %d", device, disconnected);
+
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		return -EINVAL;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	}
+
+	if (device->disconnected == disconnected)
+		return -EALREADY;
+
+	device->disconnected = disconnected;
+
+	return 0;
+}
+
 /**
  * connman_device_set_string:
  * @device: device structure
@@ -1097,7 +1572,10 @@ int connman_device_set_string(struct connman_device *device,
 {
 	DBG("device %p key %s value %s", device, key, value);
 
-	if (g_str_equal(key, "Name") == TRUE) {
+	if (g_str_equal(key, "Address") == TRUE) {
+		g_free(device->address);
+		device->address = g_strdup(value);
+	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(device->name);
 		device->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
@@ -1105,7 +1583,7 @@ int connman_device_set_string(struct connman_device *device,
 		device->node = g_strdup(value);
 	}
 
-	return 0;
+	return connman_element_set_string(&device->element, key, value);
 }
 
 /**
@@ -1120,18 +1598,20 @@ const char *connman_device_get_string(struct connman_device *device,
 {
 	DBG("device %p key %s", device, key);
 
-	if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Address") == TRUE)
+		return device->address;
+	else if (g_str_equal(key, "Name") == TRUE)
 		return device->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return device->node;
 
-	return NULL;
+	return connman_element_get_string(&device->element, key);
 }
 
-static void set_flightmode(struct connman_element *element, gpointer user_data)
+static void set_offlinemode(struct connman_element *element, gpointer user_data)
 {
 	struct connman_device *device = element->device;
-	connman_bool_t flightmode = GPOINTER_TO_UINT(user_data);
+	connman_bool_t offlinemode = GPOINTER_TO_UINT(user_data);
 	connman_bool_t powered;
 
 	DBG("element %p name %s", element, element->name);
@@ -1139,7 +1619,7 @@ static void set_flightmode(struct connman_element *element, gpointer user_data)
 	if (device == NULL)
 		return;
 
-	powered = (flightmode == TRUE) ? FALSE : TRUE;
+	powered = (offlinemode == TRUE) ? FALSE : TRUE;
 
 	if (device->powered == powered)
 		return;
@@ -1147,12 +1627,14 @@ static void set_flightmode(struct connman_element *element, gpointer user_data)
 	set_powered(device, powered);
 }
 
-int __connman_device_set_flightmode(connman_bool_t flightmode)
+int __connman_device_set_offlinemode(connman_bool_t offlinemode)
 {
-	DBG("flightmode %d", flightmode);
+	DBG("offlinmode %d", offlinemode);
 
 	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
-				set_flightmode, GUINT_TO_POINTER(flightmode));
+			set_offlinemode, GUINT_TO_POINTER(offlinemode));
+
+	__connman_notifier_offline_mode(offlinemode);
 
 	return 0;
 }
@@ -1240,6 +1722,20 @@ int connman_device_remove_network(struct connman_device *device,
 	return 0;
 }
 
+void __connman_device_set_network(struct connman_device *device,
+					struct connman_network *network)
+{
+	const char *name;
+
+	if (network != NULL) {
+		name = connman_network_get_string(network,
+						CONNMAN_PROPERTY_ID_NAME);
+		device->last_network = g_strdup(name);
+	}
+
+	device->network = network;
+}
+
 /**
  * connman_device_register:
  * @device: device structure
@@ -1408,6 +1904,26 @@ static int device_load(struct connman_device *device)
 	if (val > 0)
 		device->priority = val;
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		val = g_key_file_get_integer(keyfile, "Configuration",
+							"ScanInterval", NULL);
+		if (val > 0)
+			device->scan_interval = val;
+		break;
+	}
+
+#if 0
+	str = g_key_file_get_string(keyfile, "Configuration",
+							"LastNetwork", NULL);
+	if (str != NULL)
+		device->last_network = str;
+#endif
+
 	g_key_file_free(keyfile);
 
 	return 0;
@@ -1449,9 +1965,26 @@ update:
 		g_key_file_set_integer(keyfile, "Configuration",
 						"Priority", device->priority);
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		if (device->scan_interval > 0)
+			g_key_file_set_integer(keyfile, "Configuration",
+					"ScanInterval", device->scan_interval);
+		break;
+	}
+
+	if (device->last_network != NULL)
+		g_key_file_set_string(keyfile, "Configuration",
+					"LastNetwork", device->last_network);
+
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
-	g_file_set_contents(pathname, data, length, NULL);
+	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
+		connman_error("Failed to store device information");
 
 done:
 	g_free(data);
@@ -1460,6 +1993,9 @@ done:
 
 	g_free(pathname);
 
+	if (device->network != NULL)
+		__connman_storage_save_network(device->network);
+
 	return 0;
 }
 
